@@ -5,23 +5,30 @@ module Termonad.Term where
 import Termonad.Prelude
 
 import Control.Lens ((^.), (&), (.~), set, to)
-import Data.Colour.SRGB (RGB(RGB), toSRGB)
+import Data.Colour.SRGB (Colour, RGB(RGB), toSRGB)
+import Data.FocusList (appendFL, deleteFL, getFocusItemFL)
 import GI.Gdk
-  ( EventKey
+  ( EventButton
+  , EventKey
   , RGBA
+  , getEventButtonButton
   , newZeroRGBA
   , setRGBABlue
   , setRGBAGreen
   , setRGBARed
   )
+import GI.Gdk.Constants (pattern BUTTON_SECONDARY)
 import GI.Gio
-  ( noCancellable
+  ( menuAppend
+  , menuNew
+  , noCancellable
   )
 import GI.GLib
   ( SpawnFlags(SpawnFlagsDefault)
   )
 import GI.Gtk
   ( Align(AlignFill)
+  , ApplicationWindow
   , Box
   , Button
   , IconSize(IconSizeMenu)
@@ -45,13 +52,19 @@ import GI.Gtk
   , labelSetEllipsize
   , labelSetLabel
   , labelSetMaxWidthChars
+  , menuAttachToWidget
+  , menuNewFromModel
+  , menuPopupAtPointer
   , noAdjustment
   , notebookAppendPage
   , notebookDetachTab
+  , notebookGetNPages
   , notebookPageNum
   , notebookSetCurrentPage
+  , notebookSetShowTabs
   , notebookSetTabReorderable
   , onButtonClicked
+  , onWidgetButtonPressEvent
   , onWidgetKeyPressEvent
   , scrolledWindowNew
   , scrolledWindowSetPolicy
@@ -62,48 +75,60 @@ import GI.Gtk
   , widgetSetHalign
   , widgetSetHexpand
   , widgetShow
+  , windowSetFocus
   , windowSetTransientFor
   )
-import GI.Pango (EllipsizeMode(EllipsizeModeMiddle))
+import GI.Pango (EllipsizeMode(EllipsizeModeMiddle), FontDescription)
 import GI.Vte
-  ( CursorBlinkMode(CursorBlinkModeOn)
-  , PtyFlags(PtyFlagsDefault)
+  ( PtyFlags(PtyFlagsDefault)
   , Terminal
   , onTerminalChildExited
   , onTerminalWindowTitleChanged
   , terminalGetWindowTitle
   , terminalNew
   , terminalSetCursorBlinkMode
-  , terminalSetColorCursor
   , terminalSetFont
   , terminalSetScrollbackLines
   , terminalSpawnSync
+  , terminalSetWordCharExceptions
   )
 import System.FilePath ((</>))
 import System.Directory (getSymbolicLinkTarget)
+import System.Environment (lookupEnv)
 
-import Termonad.Config (ShowScrollbar(..), TMConfig(cursorColor, scrollbackLen), lensShowScrollbar)
-import Termonad.FocusList (appendFL, deleteFL, getFLFocusItem)
-import Termonad.Types
-  ( TMNotebookTab
-  , TMState
-  , TMState'(TMState, tmStateConfig, tmStateFontDesc, tmStateNotebook)
-  , TMTerm
-  , createTMNotebookTab
-  , lensTerm
+import Termonad.Lenses
+  ( lensConfirmExit
+  , lensOptions
+  , lensShowScrollbar
+  , lensShowTabBar
   , lensTMNotebookTabLabel
-  , lensTMNotebookTabs
   , lensTMNotebookTabTerm
   , lensTMNotebookTabTermContainer
+  , lensTMNotebookTabs
   , lensTMStateApp
   , lensTMStateConfig
   , lensTMStateNotebook
+  , lensTerm
+  )
+import Termonad.Types
+  ( ConfigHooks(createTermHook)
+  , ConfigOptions(scrollbackLen, wordCharExceptions, cursorBlinkMode)
+  , ShowScrollbar(..)
+  , ShowTabBar(..)
+  , TMConfig(hooks, options)
+  , TMNotebook
+  , TMNotebookTab
+  , TMState
+  , TMState'(TMState, tmStateAppWin, tmStateConfig, tmStateFontDesc, tmStateNotebook)
+  , TMTerm
+  , assertInvariantTMState
+  , createTMNotebookTab
   , newTMTerm
   , pid
   , tmNotebook
-  , tmNotebookTabs
   , tmNotebookTabTerm
   , tmNotebookTabTermContainer
+  , tmNotebookTabs
   )
 
 focusTerm :: Int -> TMState -> IO ()
@@ -118,10 +143,17 @@ termExitFocused :: TMState -> IO ()
 termExitFocused mvarTMState = do
   tmState <- readMVar mvarTMState
   let maybeTab =
-        tmState ^. lensTMStateNotebook . lensTMNotebookTabs . to getFLFocusItem
+        tmState ^. lensTMStateNotebook . lensTMNotebookTabs . to getFocusItemFL
   case maybeTab of
     Nothing -> pure ()
-    Just tab -> termExitWithConfirmation tab mvarTMState
+    Just tab -> termClose tab mvarTMState
+
+termClose :: TMNotebookTab -> TMState -> IO ()
+termClose tab mvarTMState = do
+  tmState <- readMVar mvarTMState
+  let confirm = tmState ^. lensTMStateConfig . lensOptions . lensConfirmExit
+      close = if confirm then termExitWithConfirmation else termExit
+  close tab mvarTMState
 
 termExitWithConfirmation :: TMNotebookTab -> TMState -> IO ()
 termExitWithConfirmation tab mvarTMState = do
@@ -181,12 +213,36 @@ relabelTabs mvarTMState = do
           term' = tmNotebookTab ^. lensTMNotebookTabTerm . lensTerm
       relabelTab notebook label scrolledWin term'
 
+-- | Compute the text for a 'Label' for a GTK Notebook tab.
+--
+-- >>> computeTabLabel 0 (Just "me@machine:~")
+-- "1. me@machine:~"
+--
+-- >>> computeTabLabel 5 (Just "bash process")
+-- "6. bash process"
+--
+-- >>> computeTabLabel 2 Nothing
+-- "3. shell"
+computeTabLabel
+  :: Int
+  -- ^ Tab number.  0 is used for the first tab, 1 for the second, etc.
+  -> Maybe Text
+  -- ^ A possible title for a tab.  If this is 'Nothing', then the string
+  -- @shell@ will be used.
+  -> Text
+computeTabLabel pageNum maybeTitle =
+  let title = fromMaybe "shell" maybeTitle
+  in tshow (pageNum + 1) <> ". " <> title
+
+-- | Update the given 'Label' for a GTK Notebook tab.
+--
+-- The new text for the label is determined by the 'computeTabLabel' function.
 relabelTab :: Notebook -> Label -> ScrolledWindow -> Terminal -> IO ()
 relabelTab notebook label scrolledWin term' = do
-  pageNum <- notebookPageNum notebook scrolledWin
+  tabNum <- notebookPageNum notebook scrolledWin
   maybeTitle <- terminalGetWindowTitle term'
-  let title = fromMaybe "bash" maybeTitle
-  labelSetLabel label $ tshow (pageNum + 1) <> ". " <> title
+  let labelText = computeTabLabel (fromIntegral tabNum) maybeTitle
+  labelSetLabel label labelText
 
 showScrollbarToPolicy :: ShowScrollbar -> PolicyType
 showScrollbarToPolicy ShowScrollbarNever = PolicyTypeNever
@@ -196,7 +252,8 @@ showScrollbarToPolicy ShowScrollbarAlways = PolicyTypeAlways
 createScrolledWin :: TMState -> IO ScrolledWindow
 createScrolledWin mvarTMState = do
   tmState <- readMVar mvarTMState
-  let showScrollbarVal = tmState ^. lensTMStateConfig . lensShowScrollbar
+  let showScrollbarVal =
+        tmState ^. lensTMStateConfig . lensOptions . lensShowScrollbar
       vScrollbarPolicy = showScrollbarToPolicy showScrollbarVal
   scrolledWin <- scrolledWindowNew noAdjustment noAdjustment
   widgetShow scrolledWin
@@ -226,10 +283,19 @@ createNotebookTabLabel = do
   widgetShow button
   pure (box, label, button)
 
-getCursorColor :: TMConfig -> IO RGBA
-getCursorColor tmConfig = do
-  let color = cursorColor tmConfig
-      RGB red green blue = toSRGB color
+setShowTabs :: TMConfig -> Notebook -> IO ()
+setShowTabs tmConfig note = do
+  npages <- notebookGetNPages note
+  let shouldShowTabs =
+        case tmConfig ^. lensOptions . lensShowTabBar of
+          ShowTabBarIfNeeded -> npages > 1
+          ShowTabBarAlways   -> True
+          ShowTabBarNever    -> False
+  notebookSetShowTabs note shouldShowTabs
+
+toRGBA :: Colour Double -> IO RGBA
+toRGBA colour = do
+  let RGB red green blue = toSRGB colour
   rgba <- newZeroRGBA
   setRGBARed rgba red
   setRGBAGreen rgba green
@@ -256,58 +322,163 @@ cwdOfPid pd = do
 #endif
 #endif
 
+-- | Get the current working directory from the shell in the focused tab of a
+-- notebook.
+--
+-- Returns 'Nothing' if there is no focused tab of the notebook, or the
+-- current working directory could not be detected for the shell.
+getCWDFromFocusedTab :: TMNotebook -> IO (Maybe Text)
+getCWDFromFocusedTab currNote = do
+  let maybeFocusedTab = getFocusItemFL (tmNotebookTabs currNote)
+  case maybeFocusedTab of
+    Nothing -> pure Nothing
+    Just focusedNotebookTab -> do
+      let shellPid = pid (tmNotebookTabTerm focusedNotebookTab)
+      cwdOfPid shellPid
 
-createTerm :: (TMState -> EventKey -> IO Bool) -> TMState -> IO TMTerm
-createTerm handleKeyPress mvarTMState = do
-  scrolledWin <- createScrolledWin mvarTMState
-  TMState{tmStateFontDesc, tmStateConfig, tmStateNotebook=currNote} <- readMVar mvarTMState
-  let maybeCurrFocusedTabPid = pid . tmNotebookTabTerm <$> getFLFocusItem (tmNotebookTabs currNote)
-  maybeCurrDir <- maybe (pure Nothing) cwdOfPid maybeCurrFocusedTabPid
+-- | Create the VTE 'Terminal', set the fonts and options
+createAndInitVteTerm :: FontDescription -> ConfigOptions -> IO Terminal
+createAndInitVteTerm tmStateFontDesc curOpts = do
   vteTerm <- terminalNew
   terminalSetFont vteTerm (Just tmStateFontDesc)
-  terminalSetScrollbackLines vteTerm (fromIntegral (scrollbackLen tmStateConfig))
-  cursorColor <- getCursorColor tmStateConfig
-  terminalSetColorCursor vteTerm (Just cursorColor)
-  terminalSetCursorBlinkMode vteTerm CursorBlinkModeOn
+  terminalSetWordCharExceptions vteTerm $ wordCharExceptions curOpts
+  terminalSetScrollbackLines vteTerm (fromIntegral (scrollbackLen curOpts))
+  terminalSetCursorBlinkMode vteTerm (cursorBlinkMode curOpts)
   widgetShow vteTerm
-  widgetGrabFocus $ vteTerm
-  terminalProcPid <-
+  pure vteTerm
+
+-- | Starts a shell in a terminal and return a new TMTerm
+launchShell
+  :: Terminal
+  -- ^ GTK 'Terminal' to spawn the shell in.
+  -> Maybe Text
+  -- ^ An optional path to the current working directory to start the
+  -- shell in.  If 'Nothing', use the current working directory of the
+  -- termonad process.
+  -> IO Int
+launchShell vteTerm maybeCurrDir = do
+  -- Should probably use GI.Vte.Functions.getUserShell, but contrary to its
+  -- documentation it raises an exception rather wrap in Maybe.
+  mShell <- lookupEnv "SHELL"
+  let argv = fromMaybe ["/usr/bin/env", "bash"] (pure <$> mShell)
+  -- Launch the shell
+  shellPid <-
     terminalSpawnSync
       vteTerm
       [PtyFlagsDefault]
       maybeCurrDir
-      ["/usr/bin/env", "bash"]
+      argv
       Nothing
       ([SpawnFlagsDefault] :: [SpawnFlags])
       Nothing
       noCancellable
-  tmTerm <- newTMTerm vteTerm (fromIntegral terminalProcPid)
-  containerAdd scrolledWin vteTerm
-  (tabLabelBox, tabLabel, tabCloseButton) <- createNotebookTabLabel
-  let notebookTab = createTMNotebookTab tabLabel scrolledWin tmTerm
-  void $
-    onButtonClicked tabCloseButton $
-      termExitWithConfirmation notebookTab mvarTMState
-  setCurrPageAction <-
-    modifyMVar mvarTMState $ \tmState -> do
+  pure (fromIntegral shellPid)
+
+-- | Add a page to the notebook and switch to it.
+addPage
+  :: TMState
+  -> TMNotebookTab
+  -> Box
+  -- ^ The GTK Object holding the label we want to show for the tab of the
+  -- newly created page of the notebook.
+  -> IO ()
+addPage mvarTMState notebookTab tabLabelBox = do
+  -- Append a new notebook page and update the TMState to reflect this.
+  (note, pageIndex) <- modifyMVar mvarTMState appendNotebookPage
+
+  -- Switch the current Notebook page to the the newly added page.
+  notebookSetCurrentPage note pageIndex
+  where
+    appendNotebookPage :: TMState' -> IO (TMState', (Notebook, Int32))
+    appendNotebookPage tmState = do
       let notebook = tmStateNotebook tmState
           note = tmNotebook notebook
           tabs = tmNotebookTabs notebook
+          scrolledWin = tmNotebookTabTermContainer notebookTab
       pageIndex <- notebookAppendPage note scrolledWin (Just tabLabelBox)
       notebookSetTabReorderable note scrolledWin True
+      setShowTabs (tmState ^. lensTMStateConfig) note
       let newTabs = appendFL tabs notebookTab
           newTMState =
             tmState & lensTMStateNotebook . lensTMNotebookTabs .~ newTabs
-          setCurrPageAction = do
-            notebookSetCurrentPage note pageIndex
-      pure (newTMState, setCurrPageAction)
-  setCurrPageAction
+      pure (newTMState, (note, pageIndex))
+
+-- | Set the keyboard focus on a vte terminal
+setFocusOn :: ApplicationWindow -> Terminal -> IO()
+setFocusOn tmStateAppWin vteTerm = do
+  widgetGrabFocus vteTerm
+  windowSetFocus tmStateAppWin (Just vteTerm)
+
+-- | Create a new 'TMTerm', setting it up and adding it to the GTKNotebook.
+createTerm
+  :: (TMState -> EventKey -> IO Bool)
+  -- ^ Funtion for handling key presses on the terminal.
+  -> TMState
+  -> IO TMTerm
+createTerm handleKeyPress mvarTMState = do
+  -- Check preconditions
+  assertInvariantTMState mvarTMState
+
+  -- Read needed data in TMVar
+  TMState{tmStateAppWin, tmStateFontDesc, tmStateConfig, tmStateNotebook=currNote} <-
+    readMVar mvarTMState
+
+  -- Create a new terminal and launch a shell in it
+  vteTerm <- createAndInitVteTerm tmStateFontDesc (options tmStateConfig)
+  maybeCurrDir <- getCWDFromFocusedTab currNote
+  termShellPid <- launchShell vteTerm maybeCurrDir
+  tmTerm <- newTMTerm vteTerm termShellPid
+
+  -- Create the container add the VTE term in it
+  scrolledWin <- createScrolledWin mvarTMState
+  containerAdd scrolledWin vteTerm
+
+  -- Create the GTK widget for the Notebook tab
+  (tabLabelBox, tabLabel, tabCloseButton) <- createNotebookTabLabel
+
+  -- Create notebook state
+  let notebookTab = createTMNotebookTab tabLabel scrolledWin tmTerm
+
+  -- Add the new notebooktab to the notebook.
+  addPage mvarTMState notebookTab tabLabelBox
+
+  -- Setup the initial label for the notebook tab.  This needs to happen
+  -- after we add the new page to the notebook, so that the page can get labelled
+  -- appropriately.
   relabelTab (tmNotebook currNote) tabLabel scrolledWin vteTerm
+
+  -- Connect callbacks
+  void $ onButtonClicked tabCloseButton $ termClose notebookTab mvarTMState
   void $ onTerminalWindowTitleChanged vteTerm $ do
     TMState{tmStateNotebook} <- readMVar mvarTMState
     let notebook = tmNotebook tmStateNotebook
     relabelTab notebook tabLabel scrolledWin vteTerm
   void $ onWidgetKeyPressEvent vteTerm $ handleKeyPress mvarTMState
   void $ onWidgetKeyPressEvent scrolledWin $ handleKeyPress mvarTMState
+  void $ onWidgetButtonPressEvent vteTerm $ handleMousePress vteTerm
   void $ onTerminalChildExited vteTerm $ \_ -> termExit notebookTab mvarTMState
+
+  -- Put the keyboard focus on the term
+  setFocusOn tmStateAppWin vteTerm
+
+  -- Make sure the state is still right
+  assertInvariantTMState mvarTMState
+
+  -- Run user-defined hooks for modifying the newly-created VTE Terminal.
+  createTermHook (hooks tmStateConfig) mvarTMState vteTerm
   pure tmTerm
+
+-- | Popup the context menu on right click
+handleMousePress :: Terminal -> EventButton -> IO Bool
+handleMousePress vteTerm event = do
+  button <- getEventButtonButton event
+  let rightClick = button == fromIntegral BUTTON_SECONDARY
+  when rightClick $ do
+    menuModel <- menuNew
+    menuAppend menuModel (Just "Copy") (Just "app.copy")
+    menuAppend menuModel (Just "Paste") (Just "app.paste")
+    menu <- menuNewFromModel menuModel
+    menuAttachToWidget menu vteTerm Nothing
+    menuPopupAtPointer menu Nothing
+  pure rightClick
+
